@@ -241,8 +241,11 @@ static void *pen_realloc(void *p, size_t n)
 
 static char *pen_strdup(char *p)
 {
-	char *b = pen_malloc(strlen(p)+1);
-	return strcpy(b, p);
+	size_t len = strlen(p);
+	char *b = pen_malloc(len+1);
+	memcpy(b, p, len);
+	b[len] = '\0';
+	return b;
 }
 
 static int pen_strncasecmp(const char *p, const char *q, size_t n)
@@ -764,20 +767,22 @@ static void setipaddress(struct in_addr *a, char *p)
 }
 
 static void setaddress(struct in_addr *a, int *port, char *s,
-		int dp, int *maxc, int *hard)
+		int dp, int *maxc, int *hard, int *weight)
 {
 	struct hostent *h;
 	char address[1024], pno[100];
-	int n = sscanf(s, "%999[^:]:%99[^:]:%d:%d", address, pno, maxc, hard);
+	int n = sscanf(s, "%999[^:]:%99[^:]:%d:%d:%d",
+		address, pno, maxc, hard, weight);
 
 	if (n > 1) *port = getport(pno);
 	else *port = dp;
 	if (n < 3) *maxc = 0;
 	if (n < 4) *hard = 0;
+	if (n < 5) *weight = 0;
 
 	if (debuglevel)
-		debug("n = %d, address = %s, pno = %d, maxc1 = %d, hard = %d",
-			n, address, *port, *maxc, *hard);
+		debug("n = %d, address = %s, pno = %d, maxc1 = %d, hard = %d, weight = %d",
+			n, address, *port, *maxc, *hard, *weight);
 
 	if (!(h = gethostbyname(address))) {
 		if ((a->s_addr = inet_addr(address)) == -1) {
@@ -1112,7 +1117,7 @@ static void usage(void)
 	       "	  [-E certfile] [-K keyfile] \\\n"
 	       "	  [-G cacertfile] [-A cacertdir] \\\n"
 	       "	  [-Z] [-R] [-L protocol] \\\n"
-	       "	  [host:]port h1[:p1[:maxc1[:hard1]]] [h2[:p2[:maxc2[:hard2]]]] ...\n"
+	       "	  [host:]port h1[:p1[:maxc1[:hard1[:weight1]]]] [h2[:p2[:maxc2[:hard2[:weight2]]]]] ...\n"
 	       "\n"
 	       "  -B host:port abuse server for naughty clients\n"
 	       "  -C port   control port\n"
@@ -1184,6 +1189,12 @@ static void init(int argc, char **argv)
 	clients = pen_calloc(clients_max, sizeof *clients);
 	/* one extra server slot for the emergency server */
 	/* and one for the abuse server */
+	/* Check that servers_max is big enough for the command line */
+	if ((argc-1) > servers_max) {
+		debug("command line specifies %d servers, max is %d; attempting to compensate",
+			argc-1, servers_max);
+		servers_max = argc-1;
+	}
 	servers = pen_calloc(servers_max+2, sizeof *servers);
 
 	nservers = 0;
@@ -1196,7 +1207,8 @@ static void init(int argc, char **argv)
 		servers[server].c = 0;	/* connections... */
 		setaddress(&servers[server].addr, &servers[server].port,
 			argv[i], port,
-			&servers[server].maxc, &servers[server].hard);
+			&servers[server].maxc, &servers[server].hard,
+			&servers[server].weight);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 
@@ -1208,7 +1220,8 @@ static void init(int argc, char **argv)
 		servers[server].c = 0;	/* connections... */
 		setaddress(&servers[server].addr, &servers[server].port,
 			"0.0.0.0", 0,
-			&servers[server].maxc, &servers[server].hard);
+			&servers[server].maxc, &servers[server].hard,
+			&servers[server].weight);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 
@@ -1221,7 +1234,8 @@ static void init(int argc, char **argv)
 		servers[server].c = 0;	/* connections... */
 		setaddress(&servers[server].addr, &servers[server].port,
 			   e_server, port,
-			   &servers[server].maxc, &servers[server].hard);
+			   &servers[server].maxc, &servers[server].hard,
+			   &servers[server].weight);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 		server++;
@@ -1233,7 +1247,8 @@ static void init(int argc, char **argv)
 		servers[server].c = 0;	/* connections... */
 		setaddress(&servers[server].addr, &servers[server].port,
 			   a_server, port,
-			   &servers[server].maxc, &servers[server].hard);
+			   &servers[server].maxc, &servers[server].hard,
+			   &servers[server].weight);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 		server++;
@@ -1629,9 +1644,10 @@ static void do_cmd(char *b, void (*output)(char *, void *), void *op)
 		}
 	} else if (!strcmp(p, "servers")) {
 		for (n = 0; n < nservers; n++) {
-			sprintf(b, "%d addr %s port %d conn %d max %d hard %d sx %llu rx %llu\n",
+			sprintf(b, "%d addr %s port %d conn %d max %d hard %d weight %d sx %llu rx %llu\n",
 				n, inet_ntoa(servers[n].addr), servers[n].port,
 				servers[n].c, servers[n].maxc, servers[n].hard,
+				servers[n].weight,
 				servers[n].sx, servers[n].rx);
 			output(b, op);
 		}
